"""
A formulation in original variables of a wedding seating problem

Authors: Stuart Mitchell 2010
"""
from time import time

import coinor.pulp as pulp
import coinor.dippy as dippy
from random import shuffle, seed
#for repeatability
seed(0)

def happiness(guest_a, guest_b):
    """
    Return the happiness (0 is the best) of allocating two
    guests together in the same table
    """
    return abs(ord(guest_a) - ord(guest_b))

def solve_problem(guests, max_table_size, max_tables):
    #time this application
    start_time = time()
    #create the set of possible tables
    tables = range(max_tables)

    possible_seatings = [(g, t) for g in guests
                                for t in tables]

    #create a binary variable to model if a guest sits at a particular table
    x = pulp.LpVariable.dicts('possible_seatings', possible_seatings, 
                                lowBound = 0,
                                upBound = 1,
                                cat = pulp.LpInteger)

    seating_model = dippy.DipProblem("Wedding Seating Model (DIP)", pulp.LpMinimize)

    #specify the maximum number of guests per table
    for table in tables:
        seating_model.relaxation[table] += sum([x[(guest, table)]
                                                for guest in guests]) <= \
                                                max_table_size, \
                                                "Maximum_table_size_%s"%table

    #A guest must seated at one and only one table
    for guest in guests:
        seating_model += sum([x[(guest, table)] for table in tables]) == 1, "Must_seat_%s"%guest

    #create a set of variables to model the objective function
    possible_pairs = [(a, b) for a in guests for b in guests if ord(a) < ord(b)]
    happy = pulp.LpVariable.dicts('table_happiness', tables,
                                  lowBound = 0,
                                  upBound = None,
                                  cat = pulp.LpContinuous)

    seating_model += sum([happy[table] for table in tables])

    #create constraints for each possible pair
    for table in tables:
        for (a, b) in possible_pairs:
            seating_model.relaxation[table] += \
                happy[table] >= happiness(a, b) * (x[(a, table)] + x[(b, table)] - 1)

    def relaxed_solver(prob, table, redCosts, convexDual):
        """
        Generate columns (tables) with negative reduced costs
        This is incorrect at the moment as the rc calculation is incorrect
        """
##        print redCosts
##        return []
        dvs = []
        neg_guests = [g for g in guests
                           if redCosts[x[(g,table)]] < 0.0]
        neg_guests.sort()
##        print neg_guests, [(i, j) for i in range(len(neg_guests))
##                                for j in range(len(neg_guests))
##                                    if j > i]
        # find all possible tables between two end points
        for pos1, pos2 in [(i, j) for i in range(len(neg_guests))
                                for j in range(len(neg_guests))
                                    if j > i]:
            # find the suitable guests that can be included in between the end points
            candidate_guests = [(redCosts[x[(g,table)]], g)
                                        for g in neg_guests[pos1+1:pos2]]
            candidate_guests.sort()
            # pick the best guests (ie those with the negative reduced costs)
            possible_table_inner = [g 
                                for _, g in candidate_guests[:max_table_size-2]]
            #This is the best table between the end points
            possible_table = [neg_guests[pos1]] + possible_table_inner +\
                                [neg_guests[pos2]]
##            print pos1, pos2, candidate_guests, possible_table_inner, possible_table
            # calculate the sum of the reduced costs for each of the guests
            neg_cost = sum(redCosts[x[(g, table)]] for g in possible_table)
            table_happiness = happiness(possible_table[0], possible_table[-1])
            #This calculation is incorrect :-(
            rc = neg_cost + table_happiness * redCosts[happy[table]] - convexDual
##            print table, possible_table, rc, [redCosts[x[g, table]] for g in possible_table], neg_cost, table_happiness, redCosts[happy[table]], convexDual
            if rc <= -0.000001:
                var_values = [(x[(g, table)], 1) 
                                for g in possible_table]
                var_values.append((happy[table], table_happiness))
                #not sure if I have the cost or rc correct here
                dv = dippy.DecompVar(var_values, rc, table_happiness)
##                return [dv]
                dvs.append(dv)

##        print dvs
        return dvs

    seating_model.relaxed_solver = relaxed_solver

    def initialise_cols(self):
        initialAlg = 1
##        print "Generating initial cols..."
        unallocated = [guest for guest in guests]
        shuffle(unallocated)
        bvs = []

        if initialAlg == 0:
            tabGuests = {}
            for table in tables:
                tabGuests[table] = []
                
            while len(unallocated) > 0:
                for table in tables:
                    if len(unallocated) > 0:
                        guest = unallocated.pop()
                        tabGuests[table].append(guest)

            for table in tables:
                var_values = []
                if tabGuests[table]:
                    for guest in tabGuests[table]:
                        var_values.append((x[(guest,table)], 1.0))
                    if len(tabGuests[table]) > 1:
                        tabHappiness = max(happiness(guest_a, guest_b)
                                           for guest_a in tabGuests[table]
                                           for guest_b in tabGuests[table]
                                           if ord(guest_a) < ord(guest_b))
                    else:
                        tabHappiness = 0.0
                    var_values.append((happy[table], tabHappiness))
                    dv = dippy.DecompVar(var_values, 0.0, tabHappiness)
##                    print dv.var_values, dv.cost
                    bvs.append((table, dv))

        else:
            for table in tables:
                tabGuests = []
                var_values = []
                for count in range(max_table_size):
                    if len(unallocated) > 0:
                        guest = unallocated.pop()
                        tabGuests.append(guest)
                        var_values.append((x[(guest,table)], 1.0))
                if tabGuests:
                    if len(tabGuests) > 1:
                        tabHappiness = max(happiness(guest_a, guest_b)
                                           for guest_a in tabGuests
                                           for guest_b in tabGuests
                                           if ord(guest_a) < ord(guest_b))
                    else:
                        tabHappiness = 0.0
                    var_values.append((happy[table], tabHappiness))
                    dv = dippy.DecompVar(var_values, 0.0, tabHappiness)
##                    print dv.var_values, dv.cost
                    bvs.append((table, dv))

        return bvs

    seating_model.init_vars = initialise_cols

    #seating_model.writeLP('wedding_main.lp')
    #for table in tables:
    #    seating_model.writeRelaxed(table, 'wedding_relax%s.lp' % table);
        
    dippy.Solve(seating_model, {
    ##        'LimitTime' : '900',
            'doPriceCut' : '1',
            'generateInitVars' : '1',
    ##        'TolZero': '%s' % tol,
    ##        'generateCuts': '0',
    ##        'CutDC': '0',
    ##        'CutCGL': '0',
    #        'LogLevel' : '5',
    #        'LogDebugLevel' : '5',
    #        'LogDumpModel' : '2',
    #        'LogLpLevel' : '2',
    #        'SolveRelaxAsIp': '1', # use default IP to solve subproblems
        })

    end_time = time() - start_time

    return end_time, seating_model.objective.value(), tables, x, happy
    
if __name__ == '__main__':
    end_time, obj, tables, x, happy = solve_problem(guests, max_table_size, max_tables)
    for table in tables:
        print table,
        for guest in guests:
            if x[(guest,table)].value() >= 0.99:
                print guest,
        print happy[table].value()
        print
    print 'Total solution time for dippy model with a solver %s seconds'%end_time
